@@ -10,8 +10,6 @@
 #include <errno.h>
 #include <signal.h>
 #include <sys/types.h>
-#include <sys/stat.h>
-#include <sys/file.h>
 #include <sys/wait.h>
 #include <sys/fcntl.h>	/* Solaris 2.5 requires this */
 
@@ -154,7 +152,7 @@ nottyDetach:
   }
 
   if (logfile)
-    fd = open(logfile, O_CREAT|O_WRONLY|O_APPEND, 0777);	/* stdout */
+    fd = open(logfile, O_CREAT|O_WRONLY|O_APPEND, 0666);	/* stdout */
   else
     if (dup(fd) < 0) {				/* stdout */
       error(0, errno, "dup");
