@@ -65,6 +65,7 @@ char *rcfile;		/* path name of rc file */
 char *idfile;		/* UID list file */
 int versioninfo;	/* emit only version info */
 char *user;		/* the name of the invoking user */
+char *fetchmailhost;	/* the name of the host running fetchmail */
 char *program_name;	/* the name to prefix error messages with */
 
 static char *lockfile;		/* name of lockfile */
@@ -115,6 +116,14 @@ int main (int argc, char **argv)
 	}
     }
 
+    /* we'll need this for the SMTP forwarding target and error messages */
+    if (gethostname(tmpbuf, sizeof(tmpbuf)))
+    {
+	fprintf(stderr, "fetchmail: can't determine fetchmail's host!");
+	exit(PS_IOERR);
+    }
+    fetchmailhost = xstrdup(tmpbuf);
+
     /*
      * Backward-compatibility hack.  If we're called by the name of the
      * ancestral popclient, look for .poprc.  This will actually work 
@@ -233,7 +242,7 @@ int main (int argc, char **argv)
 	if (check_only)
 	{
 	    fprintf(stderr,
-		 "fetchmail: can't check mail while another fetchmail to same host is running.");
+		 "fetchmail: can't check mail while another fetchmail to same host is running.\n");
 	    return(PS_EXCLUDE);
         }
 	else if (!implicitmode)
@@ -300,7 +309,7 @@ int main (int argc, char **argv)
 		}
 	    }
 
-	    if (!ctl->password)
+	    if (ctl->server.protocol != P_ETRN && !ctl->password)
 	    {
 		(void) sprintf(tmpbuf, "Enter password for %s@%s: ",
 			       ctl->remotename, ctl->server.names->id);
@@ -316,8 +325,12 @@ int main (int argc, char **argv)
     	openlog(program_name, LOG_PID, LOG_MAIL);
 #endif
 
-    if (poll_interval && !nodetach)
-	daemonize(logfile, termhook);
+    if (poll_interval)
+    {
+	if (!nodetach)
+	    daemonize(logfile, termhook);
+	error( 0, 0, "starting fetchmail %s.%s daemon ", RELEASE_ID, PATCHLEVEL);
+    }
 
     /* beyond here we don't want more than one fetchmail running per user */
     umask(0077);
@@ -513,7 +526,7 @@ static int load_params(int argc, char **argv, int optind)
     def_opts.server.protocol = P_AUTO;
     def_opts.server.timeout = CLIENT_TIMEOUT;
     def_opts.remotename = user;
-    def_opts.smtphost = "localhost";
+    save_str(&def_opts.smtphunt, -1, fetchmailhost);
 
     /* this builds the host list */
     if (prc_parse_file(rcfile) != 0)
@@ -595,32 +608,7 @@ static int load_params(int argc, char **argv, int optind)
 	    }
 #endif /* !HAVE_GETHOSTBYNAME || !HAVE_RES_SEARCH */
 
-	    /*
-	     * Assign SMTP leaders.  We want to allow all query blocks
-	     * sharing the same server/SMTP-host pair to use the same
-	     * SMTP connection.  To accomplish this, we initialize
-	     * each query block's leader field to point to the first
-	     * block in the list with a matching server/SMTP-host pair.
-	     *
-	     * In the typical case, there will be only one SMTP host (the
-	     * client machine) and thus just one SMTP leader (and one listener
-	     * process) through the entire poll cycle.
-	     */
-	    if (!ctl->mda)
-	    {
-		ctl->smtp_sockfp = (FILE *)NULL;
-		for (mp = querylist; mp && mp != ctl; mp = mp->next)
-		    if (!strcmp(mp->server.names->id, ctl->server.names->id)
-			&& !strcmp(mp->smtphost, ctl->smtphost))
-		    {
-			ctl->lead_smtp = mp->lead_smtp;
-			goto no_new_leader;
-		    }
-		ctl->lead_smtp = ctl;
-	    no_new_leader:;
-	    }
-
-	    /* similarly, compute server leaders for queries */
+	    /* compute server leaders for queries */
 	    for (mp = querylist; mp && mp != ctl; mp = mp->next)
 		if (strcmp(mp->server.names->id, ctl->server.names->id) == 0)
 		{
@@ -630,9 +618,20 @@ static int load_params(int argc, char **argv, int optind)
 	    ctl->server.lead_server = &(ctl->server);
 	no_new_server:;
 
-	    /* if stripcr hasn't been set, default it asccording to MDA */
-	    if (ctl->stripcr == -1)
-		ctl->stripcr = !ctl->mda;
+	    /* this code enables flags to be turned off */
+#define DEFAULT(flag, dflt)	if (flag == FLAG_TRUE)\
+	    				flag = TRUE;\
+				else if (flag == FLAG_FALSE)\
+					flag = FALSE;\
+				else\
+					flag = (dflt)
+	    DEFAULT(ctl->keep, FALSE);
+	    DEFAULT(ctl->flush, FALSE);
+	    DEFAULT(ctl->fetchall, FALSE);
+	    DEFAULT(ctl->rewrite, TRUE);
+	    DEFAULT(ctl->stripcr, (ctl->mda != (char *)NULL)); 
+	    DEFAULT(ctl->server.dns, TRUE);
+#undef DEFAULT
 
 	    /* plug in the semi-standard way of indicating a mail address */
 	    if (ctl->server.envelope == (char *)NULL)
@@ -689,7 +688,7 @@ void termhook(int sig)
     else
 	/* terminate all SMTP connections cleanly */
 	for (ctl = querylist; ctl; ctl = ctl->next)
-	    if (ctl->lead_smtp == ctl && ctl->smtp_sockfp != (FILE *)NULL)
+	    if (ctl->smtp_sockfp != (FILE *)NULL)
 		SMTP_quit(ctl->smtp_sockfp);
 
     if (!check_only)
@@ -709,6 +708,7 @@ static char *showproto(int proto)
     case P_IMAP: return("IMAP"); break;
     case P_APOP: return("APOP"); break;
     case P_RPOP: return("RPOP"); break;
+    case P_ETRN: return("ETRN"); break;
     default: return("unknown?!?"); break;
     }
 }
@@ -728,7 +728,8 @@ static int query_host(struct query *ctl)
 	time_t now;
 
 	time(&now);
-	fprintf(stderr, "Querying %s (protocol %s) at %s",
+	fprintf(stderr, "fetchmail: %s.%s querying %s (protocol %s) at %s",
+	    RELEASE_ID, PATCHLEVEL,
 	    ctl->server.names->id, showproto(ctl->server.protocol), ctime(&now));
     }
     switch (ctl->server.protocol) {
@@ -753,6 +754,8 @@ static int query_host(struct query *ctl)
     case P_IMAP:
 	return(doIMAP(ctl));
 	break;
+    case P_ETRN:
+	return(doETRN(ctl));
     default:
 	error(0, 0, "unsupported protocol selected.");
 	return(PS_PROTOCOL);
@@ -828,8 +831,8 @@ void dump_params (struct query *ctl)
 	   ctl->flush ? "" : " not",
 	   ctl->flush ? "on" : "off");
     printf("  Rewrite of server-local addresses is %sabled (--norewrite %s).\n",
-	   ctl->no_rewrite ? "dis" : "en",
-	   ctl->no_rewrite ? "on" : "off");
+	   ctl->rewrite ? "en" : "dis",
+	   ctl->rewrite ? "off" : "on");
     printf("  Carriage-return stripping is %sabled (--stripcr %s).\n",
 	   ctl->stripcr ? "en" : "dis",
 	   ctl->stripcr ? "on" : "off");
@@ -850,7 +853,14 @@ void dump_params (struct query *ctl)
     if (ctl->mda)
 	printf("  Messages will be delivered with '%s.'\n", visbuf(ctl->mda));
     else
-	printf("  Messages will be SMTP-forwarded to '%s'.\n", visbuf(ctl->smtphost));
+    {
+	struct idlist *idp;
+
+	printf("  Messages will be SMTP-forwarded to:");
+	for (idp = ctl->smtphunt; idp; idp = idp->next)
+	    printf(" %s", idp->id);
+	printf("\n");
+    }
     if (ctl->preconnect)
 	printf("  Server connection will be preinitialized with '%s.'\n", visbuf(ctl->preconnect));
     else if (outlevel == O_VERBOSE)
@@ -880,11 +890,14 @@ void dump_params (struct query *ctl)
 	}
 
 	printf("  DNS lookup for multidrop addresses is %sabled.\n",
-	       ctl->server.no_dns ? "dis" : "en",
-	       ctl->server.no_dns ? "on" : "off");
+	       ctl->server.dns ? "en" : "dis",
+	       ctl->server.dns ? "off" : "on");
 
 	if (count > 1)
-	    printf("  Envelope header is assumed to be: %s\n", ctl->server.envelope);
+	    if (ctl->server.envelope == STRING_DISABLED)
+		printf("  Envelope-address routing is disabled\n");
+	    else
+		printf("  Envelope header is assumed to be: %s\n", ctl->server.envelope);
     }
 #ifdef	linux
     if (ctl->server.interface)
@@ -894,7 +907,7 @@ void dump_params (struct query *ctl)
     if (ctl->server.monitor)
 	printf("  Polling loop will monitor %s.\n", ctl->server.monitor);
     else if (outlevel == O_VERBOSE)
-	printf("  No monitor interrface specified.\n");
+	printf("  No monitor interface specified.\n");
 #endif
 
     if (ctl->server.protocol > P_POP2)
