@@ -9,6 +9,7 @@
 #define		P_IMAP		4
 #define		P_APOP		5
 #define		P_RPOP		6
+#define		P_ETRN		7
 
 #define		KPOP_PORT	1109
 
@@ -46,6 +47,10 @@
 
 #define		SIZETICKER	1024	/* print 1 dot per this many bytes */
 
+/* we need to use zero as a flag-uninitialized value */
+#define FLAG_TRUE	2
+#define FLAG_FALSE	1
+
 struct idlist
 {
     char *id;
@@ -69,7 +74,7 @@ struct hostdata		/* shared among all user connections to given server */
     int timeout;
     char *envelope;
     int skip;
-    int no_dns;
+    int dns;
 
 #ifdef linux
     char *interface;
@@ -97,6 +102,7 @@ struct query
     char *remotename;
     char *password;
     char *mailbox;
+    struct idlist *smtphunt;
     char *smtphost;
     char *mda;
     char *preconnect;
@@ -105,7 +111,7 @@ struct query
     int keep;
     int fetchall;
     int flush;
-    int no_rewrite;
+    int rewrite;
     int stripcr;
     int limit;
     int fetchlimit;
@@ -118,7 +124,6 @@ struct query
     int active;
     int errcount;		/* count transient errors in last pass */
     struct query *next;		/* next query control block in chain */
-    struct query *lead_smtp;	/* pointer to this query's SMTP leader */
     FILE *smtp_sockfp;		/* socket descriptor for SMTP connection */
     unsigned int uid;		/* UID of user to deliver to */
     char digest [DIGESTLEN];	/* md5 digest buffer */
@@ -173,6 +178,7 @@ extern char *idfile;		/* path name of UID file */
 extern int linelimit;		/* limit # lines retrieved per site */
 extern int versioninfo;		/* emit only version info */
 extern char *user;		/* name of invoking user */
+extern char *fetchmailhost;	/* the name of the host running fetchmail */
 
 /* prototypes for globally callable functions */
 #if defined(HAVE_STDARG_H)
@@ -199,6 +205,7 @@ int do_protocol(struct query *, const struct method *);
 int doPOP2 (struct query *); 
 int doPOP3 (struct query *);
 int doIMAP (struct query *);
+int doETRN (struct query *);
 
 void reply_hack(char *, const char *);
 char *nxtaddr(const char *);
@@ -238,3 +245,5 @@ int yylex(void);
 
 #define FALSE	0
 #define TRUE	1
+
+#define STRING_DISABLED	(char *)-1
