@@ -80,7 +80,7 @@ void initialize_saved_lists(struct query *hostlist, const char *idfile)
 	while (fgets(buf, POPBUFSIZE, tmpfp) != (char *)NULL)
 	{
 	    /* possible lossage here with very old versions of sscanf(3)... */
-	    if ((st = sscanf(buf, "%[^@]@%s %s\n", host, user, id)) == 3)
+	    if ((st = sscanf(buf, "%[^@]@%s %s\n", user, host, id)) == 3)
 	    {
 		for (ctl = hostlist; ctl; ctl = ctl->next)
 		{
@@ -145,6 +145,7 @@ void save_str_pair(struct idlist **idl, const char *str1, const char *str2)
 	(*end)->val.id2 = xstrdup(str2);
     else
 	(*end)->val.id2 = (char *)NULL;
+    (*end)->next = (struct idlist *)NULL;
 }
 
 #ifdef __UNUSED__
@@ -221,7 +222,7 @@ void append_str_list(struct idlist **idl, struct idlist **nidl)
 	*idl = *nidl;
     else if ((*idl)->next == (struct idlist *)NULL)
 	(*idl)->next = *nidl;
-    else
+    else if ((*idl)->next != *nidl)
 	append_str_list(&(*idl)->next, nidl);
 }
 
